#Enquist, Ghirlanda & Lind:  The Human Evolutionary Transition
#Library  to be used with program figs_chapter_15_16.py
from random import seed,random,randrange,sample,choice,shuffle
from math import log
seed()



def make_random_graph(n_nodes,target, max_edges_per_node = False):
    G = dict()
    for u in range(n_nodes): G[u]=[]
    s=0
    while s<target:
        u,v = sample(range(n_nodes),2)
        if v in G[u]: continue
        if max_edges_per_node and len(G[u])==max_edges_per_node: continue 
        G[u].append(v)
        s+=1    
    return G


def behavior_map(G,target_rep_size):
    #This function generates a required behavior for each egde of the graph G
    B = dict()
    C = dict() #causal map, not used and not returned
    b_list = []
    for u in G:
        for v in G[u]:
            if not b_list: b_list = [i for i in range(target_rep_size)]
            b = b_list.pop()
            B[(u,v)] = b
            C[(u,b)] = v
    return B



Infinity=1000000
def Dijkstra_algoritm(G,source):
    Q = list(G.keys())    #create list of node keys Q
    D=dict.fromkeys(Q,Infinity)
    P=dict.fromkeys(Q,'*')
    D[source]=0  # Distance from source to source
    while len(Q) > 0: # Go trough all nodes
        j=Infinity+1
        for i in Q:
            if D[i] < j:
                j=D[i]
                u=i
        Q.remove(u)
        for v in G[u]:
            if v in Q:
                alt = D[u] + 1
                if alt < D[v]: # A shorter path to neighbor has been found
                    D[v] = alt
                    P[v] = u
    return (D,P)



def shortest_path(P,start,end):
    path = []
    while 1:
        path.append(end)
        if end == start:
            break
        if end in P:
            end = P[end]
        else:
            return [] # No path from start to end
    path.reverse()
    return path



def get_paths(G):
    #Returns a single shortest path for all connected nodes.
    paths = []
    for start in G.keys():
        D,P = Dijkstra_algoritm(G, start)
        for end in range(len(G)):
            if start!=end:
                path = shortest_path(P,start,end)
                if path: paths.append(path)
    return paths


def graph_summary(G):
    #This function prints a summary of the graph.
    n_nodes = len(G)
    result=[]
    for start in range(n_nodes):
        D,P = Dijkstra_algoritm(G,start)
        for j in D:
            if D[j]==Infinity: D[j]='NO'
        result.extend(D.values())
    while result.count('NO')>0: result.remove('NO')
    while result.count(0)>0: result.remove(0)
    connected = len(result)
    print('\nNodes: ',n_nodes)
    n_edges=0
    for i in G: n_edges += len(G[i])
    print('Edges: ',n_edges, ' of max ', n_nodes*(n_nodes-1),'(',"%.2f" % (100*n_edges/(n_nodes-1)/n_nodes),'%)')
    print('Erdős–Rényi threshold: ',"%.2f" % (100*log(n_nodes)/n_nodes),'%' )
    p = connected/(n_nodes*(n_nodes-1))
    print('Prop. connected: ',"%.3f" % (p),' (',connected,'of',n_nodes*(n_nodes-1),')' )
    if connected>0:
        print('Average distance between connected nodes: ',"%.3f" % (sum(result)/connected))
    else: print('no connections')





