#Enquist, Ghirlanda & Lind:  The Human Evolutionary Transition
#Program generating figures 15.2, 15.3, 16.3 and 16.5
from random import seed,random,randrange,sample,choice
import matplotlib.pyplot as plt
import Dijkstra
seed()

#-------------------------------------------------------------------------
#-----------------------------     Learners    ---------------------------
#-------------------------------------------------------------------------

class path_learner():
    def __init__(self,sim_parm):
        self.paths = dict()
        self.parm = sim_parm
    def learn(self,model):
        if random() < self.parm['p_social']:
            self.social_learning(model)
        else:
            self.individual_learning()
    def naive_entry(self):
        return 0 
    def entry_node_for_individual_learning(self):
        if len(self.paths) > 0:
            l = choice(list(self.paths.values()))
            return choice(l)
        else: return self.naive_entry()
    def individual_learning(self):
        u = self.entry_node_for_individual_learning()
        path = [u]
        while True:
            v = randrange( len(self.parm['world']) )
            if v not in self.parm['world'][u] or v in path: break
            path.append(v)
            u = v
        if len(path)>=2: self.update_memory(path)
    def social_learning(self,model):
        path = model.show_random_path()
        if path: self.update_memory(path)
    def update_memory(self,path):
        for i in range(len(path)-1):
            for j in range(i+1,len(path)):
                key = (path[i],path[j])
                self.paths[key] = path[i:j+1]
    def show_random_path(self):
        if len(self.paths)==0: return None
        else: return choice(list(self.paths.values()))
    def end_update(self): pass



class thinker(path_learner):
    def __init__(self,sim_parm):
        super().__init__(sim_parm)
        self.knowledge = dict()
    def entry_node_for_individual_learning(self):
        if len(self.knowledge) > 0:
            key = choice(list(self.knowledge.keys()))
            return choice(self.knowledge[key])
        else: return self.naive_entry()
    def update_memory(self,path):
        for i in range(len(path)-1):
            u,v = path[i],path[i+1]
            if u not in self.knowledge: self.knowledge[u] = [v]
            elif v not in self.knowledge[u]: self.knowledge[u].append(v)
    def end_update(self): #contructs all paths from knowledge after all learning completed
        for node in range( len(self.parm['world']) ): 
            if node not in self.knowledge:  self.knowledge[node]=[]
        D_paths = Dijkstra.get_paths(self.knowledge)
        for d_path in D_paths:
            path_key = (d_path[0],d_path[-1])
            self.paths[path_key] = d_path
    def get_n_knowledge(self):
        s = 0
        for key in self.knowledge:
            s += len(self.knowledge[key])
        return s



class repertoire_thinker(thinker):
    def __init__(self,sim_parm):
        super().__init__(sim_parm)
        self.repertoire = set()
    def individual_learning(self):
        u = self.entry_node_for_individual_learning()
        path = [u]
        while True:
            v = randrange( len(self.parm['world']) )
            if v not in self.parm['world'][u] or v in path: break
            required_b = self.parm['behavior_map'][(u,v)]
            if not (len(self.repertoire) and choice(list(self.repertoire))==required_b):  #KOLLA
                b = randrange(self.parm['behavior_flexibilty'])
                self.repertoire.add(b)
                if not b==required_b: break
            path.append(v)
            u = v
        if len(path)>=2: self.update_memory(path)
    def update_memory(self,path):
        super().update_memory(path)
        for i in range(len(path)-1):
            self.repertoire.add( self.parm['behavior_map'][(path[i],path[i+1])] )



#-------------------------------------------------------------------------
#----------------     Population analyses    -----------------------------
#-------------------------------------------------------------------------
def sequences_individuals(pop):
    n = 0
    for individual in pop: n += len(individual.paths)
    return n/len(pop)

def sequence_lengths(pop):
    max_length = 100
    data = []
    distribution = [0 for i in range(max_length)]
    for individual in pop:
        for path in individual.paths.values():
            distribution[len(path)-1]  += 1
            data.append(len(path)-1)
    s = 0
    for i in range(max_length): s += i*distribution[i]
    avg = s/sum(distribution)
    for i in range(max_length): distribution[i] = distribution[i]/len(pop)
    return avg,distribution,data

def sequences_pairs(pop):
    n,s = 0,0
    for i in range(len(pop)):
        A = set(pop[i].paths.keys())
        for j in range(i+1,len(pop)):
            B = set(pop[j].paths.keys())
            n +=1
            s += len(B.union(A))
    return s/n

def sequences_pop(pop):
    A = set()
    for i in range(len(pop)):  A.update( set(pop[i].paths.keys()) )
    return len(A)

def knowledge_individuals(pop):
    k=0
    for individual in pop: k += individual.get_n_knowledge()
    return k/len(pop)

def knowledge_pairs(pop):
    nodes = len(pop[0].knowledge)
    n,s = 0,0
    for i in range(len(pop)):
        for j in range(len(pop)):
            if i == j: continue
            n += 1
            knowledge_2 = dict()
            for u in range(nodes):
                A = set(pop[i].knowledge[u])
                B = set(pop[j].knowledge[u])
                knowledge_2[u] = list(B.union(A))
            s += len( Dijkstra.get_paths(knowledge_2) )
    return s/n

def knowledge_pop(pop):
    knowledge_all = dict()
    for u in range(len(pop[0].knowledge)):
        A = set()
        for i in range(len(pop)): A.update(pop[i].knowledge[u])
        knowledge_all[u] = list(A)
    return  len( Dijkstra.get_paths(knowledge_all) )

def repertoire_data(pop,sim_parm):
    tot,productive = 0,0
    for individual in pop:
        tot += len(individual.repertoire)
        productive += len( individual.repertoire & sim_parm['required_repertoire'] )
    return tot/len(pop), productive/len(pop)

def succces_in_pop(test_parm, pop):  #test for invasion
    opp = test_parm['learning_opportunities']
    if opp <= 0: return 0
    individual = test_parm['learner'](test_parm)
    for i in range(opp): individual.learn(choice(pop))
    individual.end_update()
    n_paths  =  len(individual.paths)
    return n_paths



#-------------------------------------------------------------------------
#-----------------------------    Simulation    --------------------------
#-------------------------------------------------------------------------

def simulate_evolution(sim_parm):
    n_seq = [0.1] #for figure purpose
    prod_rep = [0]
    coop_knowledge_pairs, coop_knowledge_pop = [0.1],[0.1]

    newpop = [sim_parm['learner'](sim_parm) for i in range(sim_parm['pop_size'])]
    for g in range(1,sim_parm['n_generations']+1):
        print(g)
        oldpop = newpop
        newpop = [sim_parm['learner'](sim_parm) for i in range(sim_parm['pop_size'])]
        for individual in newpop:
            for i in range(sim_parm['learning_opportunities']): individual.learn(choice(oldpop))
            individual.end_update()

        n_seq.append(sequences_individuals(newpop))
        if 'behavior_map' in sim_parm:
            tot, productive = repertoire_data(newpop,sim_parm)
            prod_rep.append(productive)
        if 'coop_knowledge_pairs' in sim_parm: coop_knowledge_pairs.append(knowledge_pairs(newpop))
        if 'coop_knowledge_pop' in sim_parm: coop_knowledge_pop.append(knowledge_pop(newpop))

    result = {'n_seq': n_seq,'pop':newpop}
    if 'behavior_map' in sim_parm: result['prod_rep'] = prod_rep
    if 'coop_knowledge_pairs' in sim_parm:    result['coop_knowledge_pairs'] = coop_knowledge_pairs
    if 'coop_knowledge_pop' in sim_parm:  result['coop_knowledge_pop'] = coop_knowledge_pop

    return result



#-------------------------------------------------------------------------
#-------------------------       Plots        ----------------------------
#-------------------------------------------------------------------------

def Figure_15_2():
    plt.figure(num='PowerCultEvol')
    nodes = 100 #100
    p_edge = 0.0465  # slightly  above Erdős–Rényi threshold
    edges = int(nodes*(nodes-1)*p_edge)
    world = Dijkstra.make_random_graph(nodes,edges)
    Dijkstra.graph_summary(world)
    sim_parm = {'world':world, 'pop_size':100, 'learning_opportunities':200, 'n_generations': 25 } #100,150,25
    labels  = ['Thinking + Social learning','Social learning','Thinking','Individual learning']
    learners = [thinker,path_learner,thinker,path_learner]
    p_soc = [0.25, 0.25, 0.0, 0.0]
    line_styles = ["-", "-." ,"--", ":"]
    endpop = []
    plt.subplot(1,2,1)
    for i in range(len(learners)):
        print(i,labels[i])
        sim_parm['learner'] = learners[i]
        sim_parm['p_social'] = p_soc[i]
        result = simulate_evolution(sim_parm)
        endpop.append(result['pop'])
        plt.plot( result['n_seq'], 'k', linestyle=line_styles[i], label=labels[i])
        plt.ylabel('Solved problems')
    plt.xlabel('Generation')
    plt.ylim(0,0.93)
    plt.yscale('log')
    plt.ylim(1,12000)
    plt.legend(loc = 'center right',frameon=False)
    plt.subplot(1,2,2)
    bins = [x-0.5 for x in range(1,12)]
    avg,distribution,data = sequence_lengths(endpop[0])
    plt.hist(data,bins=bins,density=True, edgecolor='black',histtype='step',linestyle='-', lw=2, label=labels[0])
    avg,length,data  = sequence_lengths(endpop[3])
    plt.hist(data,bins=bins,density=True,color='black',histtype='stepfilled',alpha=0.5, label=labels[3])
    plt.xlabel('Sequence length')
    plt.xlim(0,11)
    plt.xticks( [i for i in range(1,11)] )
    plt.ylabel("Frequency")
    plt.legend(loc = 'center right',frameon=False)
    plt.tight_layout()
    plt.show()


def Figure_15_2_Coop():
    plt.figure(num='CoopEvol')
    nodes = 100 #100
    p_edge = 0.0465  # slightly  above Erdős–Rényi threshold
    edges = int(nodes*(nodes-1)*p_edge)
    world = Dijkstra.make_random_graph(nodes,edges)
    Dijkstra.graph_summary(world)
    sim_parm = {'world':world,'pop_size':100, 'learning_opportunities':200, 'n_generations': 5} #100,100,25
    sim_parm['coop_knowledge_pairs'] = True
    sim_parm['coop_knowledge_pop'] = True
    sim_parm['learner'] = thinker
    sim_parm['p_social'] = 0.25
    
    result = simulate_evolution(sim_parm)
    plt.plot(result['coop_knowledge_pop'],'k',linestyle='-',label='all individuals')
    plt.plot(result['coop_knowledge_pairs'],'k', linestyle='--',label='two individuals')
    plt.plot(result['n_seq'], 'k', linestyle=':', label='one individual')
    print(result['n_seq'])
    print(result['coop_knowledge_pairs'])
    print(result['coop_knowledge_pop'])
    plt.xlabel('Generation')
    plt.ylabel('Solved problems')
    plt.title("Cooperation")
    plt.yscale('log')
    plt.ylim(1,21000)
    plt.legend(title='Problems that can be solved by',loc = 'lower center',frameon=False)
    plt.show()


def Figure_16_3():
    plt.figure(num='OriginCulture_imitation')
    nodes = 100 #100
    p_edge = 0.0465 #slightly  above Erdős–Rényi threshold 
    edges = int(nodes*(nodes-1)*p_edge)
    world = Dijkstra.make_random_graph(nodes,edges)
    Dijkstra.graph_summary(world)
    sim_parm = {'world':world, 'pop_size':100,  'n_generations': 1 ,'p_social': 0, 'learner': path_learner}
    learning_opportunities = [pow(2,i) for i in range(20)] #20
    print (learning_opportunities)
    individual_seq,pop_seq, imitate,  = [],[],[]
    length_1,length_2,length_over2, avg_length  = [],[],[],[]
    for opp in learning_opportunities:
        sim_parm['learning_opportunities'] = opp
        result = simulate_evolution(sim_parm)
        individual_seq.append(result['n_seq'][-1])
        avg, lengths, data = sequence_lengths(result['pop']) 
        length_1.append( lengths[1])
        length_2.append( lengths[2])
        length_over2.append( sum(lengths)- lengths[1] -lengths[2])
        avg_length.append(avg)
        pop_seq.append( sequences_pop(result['pop']) )
        #imitation in pop of indivudual learners
        test_parm = dict(sim_parm)
        test_parm['p_social'] = 0.25
        imitate.append( succces_in_pop(test_parm, result['pop']) )
    plt.subplot(1,2,1)
    plt.plot(learning_opportunities,individual_seq, 'k', linestyle='-', label='Individual alone')
    plt.plot(learning_opportunities,imitate, 'k',linestyle='--', label='Sequence imitation')
    plt.plot(learning_opportunities,pop_seq, 'k',linestyle=':', label='Available in the population')
    plt.ylabel('Solutions')
    plt.xlabel('Childhood  learning opportunities')
    plt.xscale('log')
    plt.xlim(1,1.5*learning_opportunities[-1])
    plt.legend(loc = 'upper left',frameon=False)
    plt.subplot(1,2,2)
    plt.plot(learning_opportunities,length_1, 'k', linestyle='-',         label='           1')
    plt.plot(learning_opportunities,length_2, 'k', linestyle='-.',        label='           2')
    plt.plot(learning_opportunities,length_over2, 'k', linestyle='--', label='          >2')
    plt.xlabel('Childhood  learning opportunities')
    plt.xscale('log')
    plt.xlim(1,1.5*learning_opportunities[-1])
    plt.ylabel('Number of sequences')
    plt.legend(title = 'Sequence length' , loc = 'upper left',frameon=False)
    plt.show()



def Figure_16_4():
    plt.figure(num='OriginCulture_worldmodel')
    nodes = 100 #100
    p_edge = 0.0465 #slightly  above Erdős–Rényi threshold
    edges = int(nodes*(nodes-1)*p_edge)
    world = Dijkstra.make_random_graph(nodes,edges)
    Dijkstra.graph_summary(world)
    sim_parm = {'world':world, 'pop_size': 100,  'n_generations': 1 ,'p_social': 0, 'learner': thinker}
    learning_opportunities = [pow(2,i) for i in range(20)]  #20
    percent_knowledge, percent_sequences,avg_length = [],[],[]
    for opp in learning_opportunities:
        sim_parm['learning_opportunities'] = opp
        result = simulate_evolution(sim_parm)
        percent_sequences.append( 100*result['n_seq'][-1]/(nodes*(nodes-1)) )
        percent_knowledge.append( 100*knowledge_individuals(result['pop'])/edges )
        avg,distribution,data = sequence_lengths(result['pop'])
        avg_length.append(avg)
    plt.subplot(1,2,1)
    plt.plot(learning_opportunities,percent_knowledge, 'k',linestyle='--', label='Knowledge')
    plt.plot(learning_opportunities,percent_sequences, 'k', linestyle='-', label='Problem solved')
    plt.xlabel('Childhood  learning opportunities')
    plt.xscale('log')
    plt.xlim(1,1.5*learning_opportunities[-1])
    plt.ylabel('% of total')
    plt.legend(loc = 'upper left',frameon=False)
    plt.subplot(1,2,2)
    plt.plot(learning_opportunities,avg_length, 'k', linestyle='-', label='Individual')
    plt.ylabel('Average sequence length')
    plt.xlabel('Childhood  learning opportunities')
    plt.xscale('log')
    plt.xlim(1,1.5*learning_opportunities[-1])
    plt.show()


#-------------------------------------------------------------------------
Figure_15_2() #Figure 15.2 - Left and middle panel
#Figure_15_2_Coop() #Figure 15.2 - Right panel based on this simulation
#Figure_16_3() #Figure 16.3 
#Figure_16_4() #Figure 16.4 




